"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.resolveDependencies = void 0;
const path = require("path");
const resolve = require("enhanced-resolve");
const fs = require("fs");
const source_map_1 = require("source-map");
const transpiler_1 = require("./transpiler");
const utils_1 = require("../utils");
const diagnostics_1 = require("./diagnostics");
const CompilerOptions_1 = require("../CompilerOptions");
const resolver = resolve.ResolverFactory.createResolver({
    extensions: [".lua"],
    enforceExtension: true,
    fileSystem: { ...new resolve.CachedInputFileSystem(fs) },
    useSyncFileSystemCalls: true,
});
function resolveDependencies(program, files, emitHost) {
    const outFiles = [...files];
    const diagnostics = [];
    const options = program.getCompilerOptions();
    // Resolve dependencies for all processed files
    for (const file of files) {
        if (options.tstlVerbose) {
            console.log(`Resolving dependencies for ${utils_1.normalizeSlashes(file.fileName)}`);
        }
        const resolutionResult = resolveFileDependencies(file, program, emitHost);
        outFiles.push(...resolutionResult.resolvedFiles);
        diagnostics.push(...resolutionResult.diagnostics);
    }
    return { resolvedFiles: outFiles, diagnostics };
}
exports.resolveDependencies = resolveDependencies;
function resolveFileDependencies(file, program, emitHost) {
    const dependencies = [];
    const diagnostics = [];
    const options = program.getCompilerOptions();
    for (const required of findRequiredPaths(file.code)) {
        // Do no resolve lualib
        if (required === "lualib_bundle") {
            continue;
        }
        // Do not resolve noResolution paths
        if (required.startsWith("@NoResolution:")) {
            const path = required.replace("@NoResolution:", "");
            replaceRequireInCode(file, required, path);
            replaceRequireInSourceMap(file, required, path);
            continue;
        }
        // Try to resolve the import starting from the directory `file` is in
        const fileDir = path.dirname(file.fileName);
        const resolvedDependency = resolveDependency(fileDir, required, program, emitHost);
        if (resolvedDependency) {
            if (options.tstlVerbose) {
                console.log(`Resolved ${required} to ${utils_1.normalizeSlashes(resolvedDependency)}`);
            }
            // Figure out resolved require path and dependency output path
            const resolvedRequire = transpiler_1.getEmitPathRelativeToOutDir(resolvedDependency, program);
            if (shouldRewriteRequires(resolvedDependency, program)) {
                replaceRequireInCode(file, required, resolvedRequire);
                replaceRequireInSourceMap(file, required, resolvedRequire);
            }
            // If dependency is not part of project, add dependency to output and resolve its dependencies recursively
            if (shouldIncludeDependency(resolvedDependency, program)) {
                // If dependency resolved successfully, read its content
                const dependencyContent = emitHost.readFile(resolvedDependency);
                if (dependencyContent === undefined) {
                    diagnostics.push(diagnostics_1.couldNotReadDependency(resolvedDependency));
                    continue;
                }
                const dependency = {
                    fileName: resolvedDependency,
                    code: dependencyContent,
                };
                const nestedDependencies = resolveFileDependencies(dependency, program, emitHost);
                dependencies.push(dependency, ...nestedDependencies.resolvedFiles);
                diagnostics.push(...nestedDependencies.diagnostics);
            }
        }
        else {
            // Could not resolve dependency, add a diagnostic and make some fallback path
            diagnostics.push(diagnostics_1.couldNotResolveRequire(required, path.relative(transpiler_1.getProjectRoot(program), file.fileName)));
            const fallbackRequire = fallbackResolve(required, transpiler_1.getSourceDir(program), fileDir);
            replaceRequireInCode(file, required, fallbackRequire);
            replaceRequireInSourceMap(file, required, fallbackRequire);
        }
    }
    return { resolvedFiles: dependencies, diagnostics };
}
function resolveDependency(fileDirectory, dependency, program, emitHost) {
    const options = program.getCompilerOptions();
    if (options.tstlVerbose) {
        console.log(`Resolving "${dependency}" from ${utils_1.normalizeSlashes(fileDirectory)}`);
    }
    // Check if file is a file in the project
    const resolvedPath = path.join(fileDirectory, dependency);
    const possibleProjectFiles = [
        resolvedPath,
        resolvedPath + ".ts",
        path.join(resolvedPath, "index.ts"),
        resolvedPath + ".tsx",
        path.join(resolvedPath, "index.tsx"), // tsx index
    ];
    for (const possibleFile of possibleProjectFiles) {
        if (isProjectFile(possibleFile, program)) {
            return possibleFile;
        }
    }
    // Check if this is a sibling of a required lua file
    const luaFilePath = path.resolve(fileDirectory, dependency + ".lua");
    if (emitHost.fileExists(luaFilePath)) {
        return luaFilePath;
    }
    // Not a TS file in our project sources, use resolver to check if we can find dependency
    try {
        const resolveResult = resolver.resolveSync({}, fileDirectory, dependency);
        if (resolveResult) {
            return resolveResult;
        }
    }
    catch (e) {
        // resolveSync errors if it fails to resolve
    }
    return undefined;
}
function shouldRewriteRequires(resolvedDependency, program) {
    return !isNodeModulesFile(resolvedDependency) || !isBuildModeLibrary(program);
}
function shouldIncludeDependency(resolvedDependency, program) {
    // Never include lua files (again) that are transpiled from project sources
    if (!hasSourceFileInProject(resolvedDependency, program)) {
        // Always include lua files not in node_modules (internal lua sources)
        if (!isNodeModulesFile(resolvedDependency)) {
            return true;
        }
        else {
            // Only include node_modules files if not in library mode
            return !isBuildModeLibrary(program);
        }
    }
    return false;
}
function isBuildModeLibrary(program) {
    return program.getCompilerOptions().buildMode === CompilerOptions_1.BuildMode.Library;
}
function findRequiredPaths(code) {
    // Find all require("<path>") paths in a lua code string
    const paths = [];
    const pattern = /require\("(.+)"\)/g;
    // eslint-disable-next-line @typescript-eslint/ban-types
    let match;
    while ((match = pattern.exec(code))) {
        paths.push(match[1]);
    }
    return paths;
}
function replaceRequireInCode(file, originalRequire, newRequire) {
    const requirePath = utils_1.formatPathToLuaPath(newRequire.replace(".lua", ""));
    file.code = file.code.replace(`require("${originalRequire}")`, `require("${requirePath}")`);
}
function replaceRequireInSourceMap(file, originalRequire, newRequire) {
    const requirePath = utils_1.formatPathToLuaPath(newRequire.replace(".lua", ""));
    if (file.sourceMapNode) {
        replaceInSourceMap(file.sourceMapNode, file.sourceMapNode, `"${originalRequire}"`, `"${requirePath}"`);
    }
}
function replaceInSourceMap(node, parent, require, resolvedRequire) {
    if ((!node.children || node.children.length === 0) && node.toString() === require) {
        parent.children = [new source_map_1.SourceNode(node.line, node.column, node.source, [resolvedRequire])];
        return true; // Stop after finding the first occurrence
    }
    if (node.children) {
        for (const c of node.children) {
            if (replaceInSourceMap(c, node, require, resolvedRequire)) {
                return true; // Occurrence found in one of the children
            }
        }
    }
    return false; // Did not find the require
}
function isNodeModulesFile(filePath) {
    return path
        .normalize(filePath)
        .split(path.sep)
        .some(p => p === "node_modules");
}
function isProjectFile(file, program) {
    return program.getSourceFile(file) !== undefined;
}
function hasSourceFileInProject(filePath, program) {
    const pathWithoutExtension = utils_1.trimExtension(filePath);
    return (isProjectFile(pathWithoutExtension + ".ts", program) ||
        isProjectFile(pathWithoutExtension + ".tsx", program) ||
        isProjectFile(pathWithoutExtension + ".json", program));
}
// Transform an import path to a lua require that is probably not correct, but can be used as fallback when regular resolution fails
function fallbackResolve(required, sourceRootDir, fileDir) {
    return utils_1.formatPathToLuaPath(path
        .normalize(path.join(path.relative(sourceRootDir, fileDir), required))
        .split(path.sep)
        .filter(s => s !== "." && s !== "..")
        .join(path.sep));
}
//# sourceMappingURL=resolve.js.map